from yacron import config


def test_mergedicts():
    assert dict(config.mergedicts({"a": 1}, {"b": 2})) == {"a": 1, "b": 2}


def test_mergedicts_nested():
    assert (dict(config.mergedicts(
        {"a": {'x': 1, 'y': 2, 'z': 3}},
        {'a': {'y': 10}, "b": 2}
    )) == {"a": {'x': 1, 'y': 10, 'z': 3}, "b": 2})


def test_mergedicts_right_none():
    assert (dict(config.mergedicts(
        {"a": {'x': 1}},
        {"a": None, "b": 2}
    )) == {"a": {'x': 1}, "b": 2})


def test_mergedicts_lists():
    assert (dict(config.mergedicts(
        {"env": [{'key': 'FOO'}]},
        {"env": [{'key': 'BAR'}]}
    )) == {"env": [{'key': 'FOO'}, {'key': 'BAR'}]})


def test_simple_config1():
    jobs = config.parse_config_string('''
defaults:
  shell: /bin/bash

jobs:
  - name: test-03
    command: |
      trap "echo '(ignoring SIGTERM)'" TERM
      echo "starting..."
      sleep 10
      echo "all done."
    schedule:
      minute: "*"
    captureStderr: true
    executionTimeout: 1
    killTimeout: 0.5
                       ''')
    assert len(jobs) == 1
    job = jobs[0]
    assert job.name == 'test-03'
    assert job.command == ('trap "echo \'(ignoring SIGTERM)\'" TERM\n'
                           'echo "starting..."\n'
                           'sleep 10\n'
                           'echo "all done."\n')
    assert job.schedule_unparsed == {'minute': '*'}
    assert job.captureStderr is True
    assert job.captureStdout is False
    assert job.executionTimeout == 1
    assert job.killTimeout == 0.5


def test_config_default_report():
    jobs = config.parse_config_string('''
defaults:
  onFailure:
    report:
      mail:
        from: example@foo.com
        to: example@bar.com
        smtpHost: 127.0.0.1
        smtpPort: 10025

jobs:
  - name: test-03
    command: foo
    schedule:
      minute: "*"
    captureStderr: true
                       ''')
    assert len(jobs) == 1
    job = jobs[0]
    assert job.onFailure == (
        {'report': {'mail': {'from': 'example@foo.com',
                             'smtpHost': '127.0.0.1',
                             'smtpPort': 10025,
                             'smtp_host': None,
                             'smtp_port': 25,
                             'to': 'example@bar.com'},
                    'sentry': {'dsn': {'fromEnvVar': None,
                                       'fromFile': None,
                                       'value': None}}},
         'retry': {'backoffMultiplier': 2,
                   'initialDelay': 1,
                   'maximumDelay': 300,
                   'maximumRetries': 0}})


def test_config_default_report_override():
    # even if the default says send email on error, it should be possible for
    # specific jobs to override the default and disable sending email.
    jobs = config.parse_config_string('''
defaults:
  onFailure:
    report:
      mail:
        from: example@foo.com
        to: example@bar.com
        smtpHost: 127.0.0.1
        smtpPort: 10025

jobs:
  - name: test-03
    command: foo
    schedule:
      minute: "*"
    captureStderr: true
    onFailure:
      report:
        mail:
          to:
          from:
                       ''')
    assert len(jobs) == 1
    job = jobs[0]
    assert job.onFailure == (
        {'report': {'mail': {'from': None,
                             'smtpHost': '127.0.0.1',
                             'smtpPort': 10025,
                             'smtp_host': None,
                             'smtp_port': 25,
                             'to': None},
                    'sentry': {'dsn': {'fromEnvVar': None,
                                       'fromFile': None,
                                       'value': None}}},
         'retry': {'backoffMultiplier': 2,
                   'initialDelay': 1,
                   'maximumDelay': 300,
                   'maximumRetries': 0}})



import time
import datetime
from pathlib import Path

import yacron.cron
from yacron.job import RunningJob
from yacron.config import JobConfig
import asyncio
import pytest


@pytest.fixture(autouse=True)
def fixed_current_time(monkeypatch):
    FIXED_TIME = datetime.datetime(year=1999, month=12, day=31, hour=12,
                                   minute=0, second=0)
    def get_now():
        return FIXED_TIME
    monkeypatch.setattr("yacron.cron.get_now", get_now)


class TracingRunningJob(RunningJob):

    _TRACE = asyncio.Queue()

    def __init__(self, config: JobConfig, retry_state) -> None:
        super().__init__(config, retry_state)
        self._TRACE.put_nowait((time.perf_counter(), "create", self))

    async def start(self) -> None:
        self._TRACE.put_nowait((time.perf_counter(), "start", self))
        await super().start()
        self._TRACE.put_nowait((time.perf_counter(), "started", self))

    async def wait(self) -> None:
        self._TRACE.put_nowait((time.perf_counter(), "wait", self))
        await super().wait()
        self._TRACE.put_nowait((time.perf_counter(), "waited", self))

    async def cancel(self) -> None:
        self._TRACE.put_nowait((time.perf_counter(), "cancel", self))
        await super().cancel()
        self._TRACE.put_nowait((time.perf_counter(), "cancelled", self))

    async def report_failure(self):
        self._TRACE.put_nowait((time.perf_counter(), "report_failure", self))
        await super().report_failure()

    async def report_permanent_failure(self):
        self._TRACE.put_nowait((time.perf_counter(),
                                "report_permanent_failure", self))
        await super().report_permanent_failure()

    async def report_success(self):
        self._TRACE.put_nowait((time.perf_counter(), "report_success", self))
        await super().report_success()


JOB_THAT_SUCCEEDS = '''
jobs:
  - name: test
    command: |
      echo "foobar"
    schedule: "* * * * *"
'''

JOB_THAT_FAILS = '''
jobs:
  - name: test
    command: |
      echo "foobar"
      exit 2
    schedule: "* * * * *"
'''


@pytest.mark.parametrize("config_yaml, expected_events", [
    (JOB_THAT_SUCCEEDS, ['create', 'start', 'started', 'wait', 'waited',
                         'report_success']),
    (JOB_THAT_FAILS, ['create', 'start', 'started', 'wait', 'waited',
                      'report_failure', 'report_permanent_failure']),
])
def test_simple(monkeypatch, config_yaml, expected_events):
    monkeypatch.setattr(yacron.cron, "RunningJob", TracingRunningJob)
    cron = yacron.cron.Cron(None, config_yaml=config_yaml)

    events = []

    async def wait_and_quit():
        the_job = None
        while True:
            ts, event, job = await TracingRunningJob._TRACE.get()
            print(ts, event)
            if the_job is None:
                job = the_job
            else:
                assert job is the_job
            events.append(event)
            if event in {'report_success', 'report_permanent_failure'}:
                break
        cron.signal_shutdown()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(asyncio.gather(
        wait_and_quit(),
        cron.run()))
    assert events == expected_events


RETRYING_JOB_THAT_FAILS = '''
jobs:
  - name: test
    command: |
      echo "foobar"
      exit 2
    schedule: "* * * * *"
    onFailure:
      retry:
        maximumRetries: 2
        initialDelay: 0.1
        maximumDelay: 1
        backoffMultiplier: 2
'''


def test_fail_retry(monkeypatch):
    monkeypatch.setattr(yacron.cron, "RunningJob", TracingRunningJob)
    cron = yacron.cron.Cron(None, config_yaml=RETRYING_JOB_THAT_FAILS)

    events = []

    async def wait_and_quit():
        known_jobs = {}
        while True:
            ts, event, job = await TracingRunningJob._TRACE.get()
            try:
                jobnum = known_jobs[job]
            except KeyError:
                if known_jobs:
                    jobnum = max(known_jobs.values()) + 1
                else:
                    jobnum = 1
                known_jobs[job] = jobnum
            print(ts, event, jobnum)
            events.append((jobnum, event))
            if jobnum == 3 and event == 'report_permanent_failure':
                break
        cron.signal_shutdown()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(asyncio.gather(
        wait_and_quit(),
        cron.run()))
    assert events == [
        # initial attempt
        (1, 'create'),
        (1, 'start'), (1, 'started'),
        (1, 'wait'), (1, 'waited'),
        (1, 'report_failure'),
        # first retry
        (2, 'create'), (2, 'start'), (2, 'started'),
        (2, 'wait'), (2, 'waited'), (2, 'report_failure'),
        # second retry
        (3, 'create'), (3, 'start'), (3, 'started'),
        (3, 'wait'), (3, 'waited'),
        (3, 'report_failure'), (3, 'report_permanent_failure')]


JOB_THAT_HANGS = '''
jobs:
  - name: test
    command: |
      trap "echo '(ignoring SIGTERM)'" TERM
      echo "starting..."
      sleep 10
      echo "all done."
    schedule: "* * * * *"
    captureStdout: true
    executionTimeout: 0.25
    killTimeout: 0.25
'''


def test_execution_timeout(monkeypatch):
    monkeypatch.setattr(yacron.cron, "RunningJob", TracingRunningJob)
    cron = yacron.cron.Cron(None, config_yaml=JOB_THAT_HANGS)

    events = []
    jobs_stdout = {}

    async def wait_and_quit():
        known_jobs = {}
        while True:
            ts, event, job = await TracingRunningJob._TRACE.get()
            try:
                jobnum = known_jobs[job]
            except KeyError:
                if known_jobs:
                    jobnum = max(known_jobs.values()) + 1
                else:
                    jobnum = 1
                known_jobs[job] = jobnum
            print(ts, event, jobnum)
            events.append((jobnum, event))
            if jobnum == 1 and event == 'report_permanent_failure':
                jobs_stdout[jobnum] = job.stdout
                break
        cron.signal_shutdown()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(asyncio.gather(
        wait_and_quit(),
        cron.run()))
    assert events == [
        # initial attempt
        (1, 'create'),
        (1, 'start'), (1, 'started'),
        (1, 'wait'),
        (1, 'cancel'), (1, 'cancelled'),
        (1, 'waited'),
        (1, 'report_failure'), (1, 'report_permanent_failure')]
    assert jobs_stdout[1] == 'starting...\n'


CONCURRENT_JOB = '''
jobs:
  - name: test
    command: |
      echo "starting..."
      sleep 0.5
      echo "all done."
    schedule: "* * * * *"
    captureStdout: true
    concurrencyPolicy: {policy}
'''


@pytest.mark.parametrize("policy,expected_numjobs,expected_max_running", [
    ('Allow', 2, 2),
    ('Forbid', 1, 1),
    ('Replace', 2, 1),
])
def test_concurrency_policy(monkeypatch, policy,
                            expected_numjobs, expected_max_running):
    monkeypatch.setattr(yacron.cron, "RunningJob", TracingRunningJob)
    START_TIME = datetime.datetime(year=1999, month=12, day=31, hour=12,
                                   minute=0, second=59, microsecond=750000)

    t0 = time.perf_counter()

    def get_now():
        return (START_TIME +
                datetime.timedelta(seconds=(time.perf_counter() - t0)))
    monkeypatch.setattr("yacron.cron.get_now", get_now)

    cron = yacron.cron.Cron(
        None,
        config_yaml=CONCURRENT_JOB.format(policy=policy),
    )

    events = []
    numjobs = 0
    max_running = 0

    async def wait_and_quit():
        nonlocal numjobs, max_running
        known_jobs = {}
        pending_jobs = set()
        running_jobs = set()
        while not (known_jobs and not pending_jobs):
            ts, event, job = await TracingRunningJob._TRACE.get()
            try:
                jobnum = known_jobs[job]
            except KeyError:
                if known_jobs:
                    jobnum = max(known_jobs.values()) + 1
                else:
                    jobnum = 1
                known_jobs[job] = jobnum
                pending_jobs.add(jobnum)
                running_jobs.add(jobnum)
                numjobs += 1
            print(ts, event, jobnum)
            events.append((jobnum, event))
            if event in {'report_success', 'report_permanent_failure'}:
                pending_jobs.discard(jobnum)
            if event in {'report_success', 'report_permanent_failure',
                         'cancelled'}:
                running_jobs.discard(jobnum)
            max_running = max(len(running_jobs), max_running)
        cron.signal_shutdown()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(asyncio.gather(
        wait_and_quit(),
        cron.run()))
    import pprint
    pprint.pprint(events)
    assert (numjobs, max_running) == (expected_numjobs, expected_max_running)


def test_simple_config_file(monkeypatch):
    monkeypatch.setattr(yacron.cron, "RunningJob", TracingRunningJob)
    config_arg = str(Path(__file__).parent / 'testconfig.yaml')
    yacron.cron.Cron(config_arg)


RETRYING_JOB_THAT_FAILS2 = '''
jobs:
  - name: test
    command: |
      echo "foobar"
      exit 2
    schedule: "* * * * *"
    onFailure:
      retry:
        maximumRetries: 1
        initialDelay: 0.4
        maximumDelay: 1
        backoffMultiplier: 1
'''
def test_concurrency_and_backoff(monkeypatch):
    monkeypatch.setattr(yacron.cron, "RunningJob", TracingRunningJob)
    START_TIME = datetime.datetime(year=1999, month=12, day=31, hour=12,
                                   minute=0, second=59, microsecond=750000)
    STOP_TIME = datetime.datetime(year=1999, month=12, day=31, hour=12,
                                  minute=1, second=00, microsecond=250000)

    t0 = time.perf_counter()

    def get_now():
        return (START_TIME +
                datetime.timedelta(seconds=(time.perf_counter() - t0)))

    def get_reltime(ts):
        return (START_TIME + datetime.timedelta(seconds=(ts - t0)))

    monkeypatch.setattr("yacron.cron.get_now", get_now)

    cron = yacron.cron.Cron(None, config_yaml=RETRYING_JOB_THAT_FAILS2)

    events = []
    numjobs = 0

    async def wait_and_quit():
        nonlocal numjobs
        known_jobs = {}
        pending_jobs = set()
        running_jobs = set()
        while get_now() < STOP_TIME:
            try:
                ts, event, job = await asyncio.wait_for(
                    TracingRunningJob._TRACE.get(), 0.1)
            except asyncio.TimeoutError:
                continue
            try:
                jobnum = known_jobs[job]
            except KeyError:
                if known_jobs:
                    jobnum = max(known_jobs.values()) + 1
                else:
                    jobnum = 1
                known_jobs[job] = jobnum
                pending_jobs.add(jobnum)
                running_jobs.add(jobnum)
                numjobs += 1
            print(get_reltime(ts), event, jobnum)
            events.append((jobnum, event))
            if event in {'report_success', 'report_permanent_failure'}:
                pending_jobs.discard(jobnum)
            if event in {'report_success', 'report_permanent_failure',
                         'cancelled'}:
                running_jobs.discard(jobnum)
        cron.signal_shutdown()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(asyncio.gather(
        wait_and_quit(),
        cron.run()))
    import pprint
    pprint.pprint(events)
    assert numjobs == 2



import logging
import os.path
from typing import Union  # noqa
from typing import List, Optional

import strictyaml
from strictyaml import Optional as Opt
from strictyaml import Bool, EmptyNone, Enum, Float, Int, Map, Seq, Str
from strictyaml.exceptions import StrictYAMLError
from ruamel.yaml.error import YAMLError

from crontab import CronTab

logger = logging.getLogger('yacron.config')


class ConfigError(Exception):
    pass


_REPORT_DEFAULTS = {
    'sentry': {
        'dsn': {
            'value': None,
            'fromFile': None,
            'fromEnvVar': None,
        },
    },
    'mail': {
        'from': None,
        'to': None,
        'smtp_host': None,  # deprecated
        'smtp_port': 25,  # deprecated
        'smtpHost': None,
        'smtpPort': 25,
    },
}


DEFAULT_CONFIG = {
    'shell': '/bin/sh',
    'concurrencyPolicy': 'Allow',
    'captureStderr': True,
    'captureStdout': False,
    'saveLimit': 4096,
    'failsWhen': {
        'producesStdout': False,
        'producesStderr': True,
        'nonzeroReturn': True,
    },
    'onFailure': {
        'retry': {
            'maximumRetries': 0,
            'initialDelay': 1,
            'maximumDelay': 300,
            'backoffMultiplier': 2,
        },
        'report': _REPORT_DEFAULTS,
    },
    'onPermanentFailure': {
        'report': _REPORT_DEFAULTS,
    },
    'onSuccess': {
        'report': _REPORT_DEFAULTS,
    },
    'environment': [],
    'executionTimeout': None,
    'killTimeout': 30,
}


_report_schema = Map({
    "sentry": Map({
        Opt("dsn"): Map({
            Opt("value"): Str() | EmptyNone(),
            Opt("fromFile"): Str() | EmptyNone(),
            Opt("fromEnvVar"): Str() | EmptyNone(),
        }),
    }),
    Opt("mail"): Map({
        "from": EmptyNone() | Str(),
        "to": EmptyNone() | Str(),
        Opt("smtp_host"): Str(),
        Opt("smtp_port"): Int(),
        Opt("smtpHost"): Str(),
        Opt("smtpPort"): Int(),
    })
})

_job_defaults_common = {
    Opt("shell"): Str(),
    Opt("concurrencyPolicy"): Enum(['Allow', 'Forbid', 'Replace']),
    Opt("captureStderr"): Bool(),
    Opt("captureStdout"): Bool(),
    Opt("saveLimit"): Int(),
    Opt("failsWhen"): Map({
        "producesStdout": Bool(),
        "producesStderr": Bool(),
        "nonzeroReturn": Bool(),
    }),
    Opt("onFailure"): Map({
        Opt("retry"): Map({
            "maximumRetries": Int(),
            "initialDelay": Float(),
            "maximumDelay": Float(),
            "backoffMultiplier": Float(),
        }),
        Opt("report"): _report_schema,
    }),
    Opt("onPermanentFailure"): Map({
        Opt("report"): _report_schema,
    }),
    Opt("onSuccess"): Map({
        Opt("report"): _report_schema,
    }),
    Opt("environment"): Seq(Map({
        "key": Str(),
        "value": Str(),
    })),
    Opt("executionTimeout"): Float(),
    Opt("killTimeout"): Float(),
}

_job_schema_dict = dict(_job_defaults_common)
_job_schema_dict.update({
    "name": Str(),
    "command": Str() | Seq(Str()),
    "schedule": Str() | Map({
        Opt("minute"): Str(),
        Opt("hour"): Str(),
        Opt("dayOfMonth"): Str(),
        Opt("month"): Str(),
        Opt("year"): Str(),
        Opt("dayOfWeek"): Str(),
    }),
})

CONFIG_SCHEMA = Map({
    Opt("defaults"): Map(_job_defaults_common),
    "jobs": Seq(Map(_job_schema_dict)),
})


# Slightly modified version of https://stackoverflow.com/a/7205672/2211825
def mergedicts(dict1, dict2):
    for k in set(dict1.keys()).union(dict2.keys()):
        if k in dict1 and k in dict2:
            v1 = dict1[k]
            v2 = dict2[k]
            if isinstance(v1, dict) and isinstance(v2, dict):
                yield (k, dict(mergedicts(v1, v2)))
            elif isinstance(v1, dict) and v2 is None:  # modification
                yield (k, dict(mergedicts(v1, {})))
            elif isinstance(v1, list) and isinstance(v2, list):  # merge lists
                yield (k, v1 + v2)
            else:
                yield (k, v2)
        elif k in dict1:
            yield (k, dict1[k])
        else:
            yield (k, dict2[k])


class JobConfig:

    def __init__(self, config: dict) -> None:
        self.name = config['name']  # type: str
        self.command = config['command']  # type: Union[str, List[str]]
        self.schedule_unparsed = config.pop('schedule')
        if isinstance(self.schedule_unparsed, str):
            self.schedule = CronTab(self.schedule_unparsed)  # type: CronTab
        elif isinstance(self.schedule_unparsed, dict):
            minute = self.schedule_unparsed.get("minute", "*")
            hour = self.schedule_unparsed.get("hour", "*")
            day = self.schedule_unparsed.get("dayOfMonth", "*")
            month = self.schedule_unparsed.get("month", "*")
            dow = self.schedule_unparsed.get("dayOfWeek", "*")
            tab = '{} {} {} {} {}'.format(minute, hour, day, month, dow)
            logger.debug("Converted schedule to %r", tab)
            self.schedule = CronTab(tab)
        else:
            raise ValueError("invalid schedule: %r", self.schedule_unparsed)
        self.shell = config.pop('shell')
        self.concurrencyPolicy = config.pop('concurrencyPolicy')
        self.captureStderr = config.pop('captureStderr')
        self.captureStdout = config.pop('captureStdout')
        self.saveLimit = config.pop('saveLimit')
        self.failsWhen = config.pop('failsWhen')
        self.onFailure = config.pop('onFailure')
        self.onPermanentFailure = config.pop('onPermanentFailure')
        self.onSuccess = config.pop('onSuccess')
        self.environment = config.pop('environment')
        self.executionTimeout = config.pop('executionTimeout')
        self.killTimeout = config.pop('killTimeout')


def parse_config_file(path: str) -> List[JobConfig]:
    with open(path, "rt", encoding='utf-8') as stream:
        data = stream.read()
    return parse_config_string(data, path)


def parse_config_string(data: str, path: Optional[str] = None,
                        ) -> List[JobConfig]:
    try:
        doc = strictyaml.load(data, CONFIG_SCHEMA).data
    except StrictYAMLError as ex:
        if ex.context_mark is not None:
            ex.context_mark.name = path
        if ex.problem_mark is not None:
            ex.problem_mark.name = path
        raise ConfigError(str(ex))
    except YAMLError as ex:
        raise ConfigError(str(ex))

    defaults = doc.get('defaults', {})
    jobs = []
    for config_job in doc['jobs']:
        job_dict = dict(mergedicts(DEFAULT_CONFIG, defaults))
        job_dict = dict(mergedicts(job_dict, config_job))
        jobs.append(JobConfig(job_dict))
    return jobs


def parse_config(config_arg: str) -> List[JobConfig]:
    jobs = []
    config_errors = {}
    if os.path.isdir(config_arg):
        for direntry in os.scandir(config_arg):
            _, ext = os.path.splitext(direntry.name)
            if ext in {'.yml', '.yaml'}:
                try:
                    config = parse_config_file(direntry.path)
                except ConfigError as err:
                    config_errors[direntry.path] = str(err)
                except OSError as ex:
                    config_errors[config_arg] = str(ex)
                else:
                    jobs.extend(config)
    else:
        try:
            config = parse_config_file(config_arg)
        except OSError as ex:
            config_errors[config_arg] = str(ex)
        else:
            jobs.extend(config)
    if config_errors:
        raise ConfigError("\n---".join(config_errors.values()))
    return jobs



import yacron.job
import yacron.config
import asyncio
import pytest
import aiosmtplib
import raven
from unittest.mock import Mock, patch


@pytest.mark.parametrize("save_limit, output", [
    (10, 'line1\nline2\nline3\nline4\n'),
    (1, '   [.... 3 lines discarded ...]\nline4\n'),
    (2, 'line1\n   [.... 2 lines discarded ...]\nline4\n'),
])
def test_stream_reader(save_limit, output):
    loop = asyncio.get_event_loop()
    fake_stream = asyncio.StreamReader()
    reader = yacron.job.StreamReader("cronjob-1", "stderr", fake_stream,
                                     save_limit)

    async def producer(fake_stream):
        fake_stream.feed_data(b"line1\nline2\nline3\nline4\n")
        fake_stream.feed_eof()

    _, out = loop.run_until_complete(asyncio.gather(
        producer(fake_stream),
        reader.join()))

    assert out == output


A_JOB = '''
jobs:
  - name: test
    command: ls
    schedule: "* * * * *"
    onSuccess:
      report:
        mail:
          from: example@foo.com
          to: example@bar.com
          smtpHost: smtp1
          smtpPort: 1025
'''


@pytest.mark.parametrize("report_type, stdout, stderr, subject, body", [
    (yacron.job.ReportType.SUCCESS, "out", "err",
     "Cron job 'test' completed",
     'STDOUT:\n---\nout\n---\nSTDERR:\nerr'),

    (yacron.job.ReportType.FAILURE, "out", "err",
     "Cron job 'test' failed",
     'STDOUT:\n---\nout\n---\nSTDERR:\nerr'),

    (yacron.job.ReportType.FAILURE, None, None,
     "Cron job 'test' failed",
     "(no output was captured)"),

    (yacron.job.ReportType.FAILURE, None, "err",
     "Cron job 'test' failed",
     'err'),

    (yacron.job.ReportType.FAILURE, "out", None,
     "Cron job 'test' failed",
     'out'),
])
def test_report_mail(report_type, stdout, stderr, subject, body):
    job_config = yacron.config.parse_config_string(A_JOB)[0]
    job = Mock(config=job_config, stdout=stdout, stderr=stderr)
    mail = yacron.job.MailReporter()
    loop = asyncio.get_event_loop()

    connect_calls = []
    messages_sent = []

    async def connect(self):
        connect_calls.append(self)

    async def send_message(self, message):
        messages_sent.append(message)

    real_init = aiosmtplib.SMTP.__init__
    smtp_init_args = None
    def init(self, *args, **kwargs):
        nonlocal smtp_init_args
        smtp_init_args = args, kwargs
        real_init(self, *args, **kwargs)

    with patch("aiosmtplib.SMTP.__init__", init), \
         patch("aiosmtplib.SMTP.connect", connect), \
         patch("aiosmtplib.SMTP.send_message", send_message):
        loop.run_until_complete(mail.report(report_type,
                                            job,
                                            job_config.onSuccess['report']))

    assert smtp_init_args == ((), {'hostname': 'smtp1', 'port': 1025})
    assert len(connect_calls) == 1
    assert len(messages_sent) == 1
    message = messages_sent[0]
    assert message['From'] == "example@foo.com"
    assert message['To'] == "example@bar.com"
    assert message['Subject'] == subject
    assert message.get_payload() == body


@pytest.mark.parametrize("report_type, dsn_from, body, extra, expected_dsn", [
    (yacron.job.ReportType.SUCCESS,
     "value",
     "Cron job 'test' completed\n\nSTDOUT:\n---\nout\n---\nSTDERR:\nerr",
     {
        'job': 'test',
        'exit_code': 0,
        'command': 'ls',
        'shell': '/bin/sh',
        'success': True,
    }, "http://xxx:yyy@sentry/1"),
    (yacron.job.ReportType.FAILURE,
     "file",
     "Cron job 'test' failed\n\nSTDOUT:\n---\nout\n---\nSTDERR:\nerr",
     {
        'job': 'test',
        'exit_code': 0,
        'command': 'ls',
        'shell': '/bin/sh',
        'success': False,
    }, "http://xxx:yyy@sentry/2"),
    (yacron.job.ReportType.FAILURE,
     "envvar",
     "Cron job 'test' failed\n\nSTDOUT:\n---\nout\n---\nSTDERR:\nerr",
     {
        'job': 'test',
        'exit_code': 0,
        'command': 'ls',
        'shell': '/bin/sh',
        'success': False,
    }, "http://xxx:yyy@sentry/3"),
])
def test_report_sentry(report_type, dsn_from, body, extra, expected_dsn,
                       tmpdir, monkeypatch):
    job_config = yacron.config.parse_config_string(A_JOB)[0]

    p = tmpdir.join("sentry-secret-dsn")
    p.write("http://xxx:yyy@sentry/2")

    monkeypatch.setenv("TEST_SENTRY_DSN", "http://xxx:yyy@sentry/3")

    if dsn_from == 'value':
        job_config.onSuccess['report']['sentry'] = {
            "dsn": {
                "value": "http://xxx:yyy@sentry/1",
                "fromFile": None,
                "fromEnvVar": None,
            }
        }
    elif dsn_from == 'file':
        job_config.onSuccess['report']['sentry'] = {
            "dsn": {
                "value": None,
                "fromFile": str(p),
                "fromEnvVar": None,
            }
        }
    elif dsn_from == 'envvar':
        job_config.onSuccess['report']['sentry'] = {
            "dsn": {
                "value": None,
                "fromFile": None,
                "fromEnvVar": "TEST_SENTRY_DSN",
            }
        }
    else:
        raise AssertionError

    job = Mock(config=job_config, stdout="out", stderr="err", retcode=0)
    loop = asyncio.get_event_loop()

    messages_sent = []

    def captureMessage(self, body, extra):
        messages_sent.append((body, extra))

    real_init = raven.Client.__init__
    init_args = (), {}

    def init(self, *args, **kwargs):
        nonlocal init_args
        init_args = args, kwargs
        real_init(self, *args, **kwargs)

    sentry = yacron.job.SentryReporter()
    with patch("raven.Client.__init__", init), \
         patch("raven.Client.captureMessage", captureMessage):
        loop.run_until_complete(sentry.report(report_type,
                                              job,
                                              job_config.onSuccess['report']))

    args, kwargs = init_args
    assert kwargs.get('dsn') == expected_dsn

    assert len(messages_sent) == 1
    got_body, got_extra = messages_sent[0]
    assert got_body == body
    assert got_extra == extra


@pytest.mark.parametrize("shell, command, expected_type, expected_args", [
    ('', "Civ 6", 'shell', ('Civ 6',)),
    ('', ["echo", "hello"], 'exec', ('echo', 'hello')),
    ('bash', 'echo "hello"', 'exec', ('bash', '-c', 'echo "hello"',)),
])
def test_job_run(monkeypatch, shell, command, expected_type, expected_args):

    shell_commands = []
    exec_commands = []

    async def create_subprocess_common(*args, **kwargs):
        stdout = asyncio.StreamReader()
        stderr = asyncio.StreamReader()
        stdout.feed_data(b"out\n")
        stdout.feed_eof()
        stderr.feed_data(b"err\n")
        stderr.feed_eof()
        proc = Mock(stdout=stdout, stderr=stderr)

        async def wait():
            return

        proc.wait = wait
        return proc

    async def create_subprocess_shell(*args, **kwargs):
        shell_commands.append((args, kwargs))
        return await create_subprocess_common(*args, **kwargs)

    async def create_subprocess_exec(*args, **kwargs):
        exec_commands.append((args, kwargs))
        return await create_subprocess_common(*args, **kwargs)

    monkeypatch.setattr("asyncio.create_subprocess_exec",
                        create_subprocess_exec)
    monkeypatch.setattr("asyncio.create_subprocess_shell",
                        create_subprocess_shell)

    if isinstance(command, list):
        command_snippet = '\n'.join(
            ["    command:"] + ['      - ' + arg for arg in command])
    else:
        command_snippet = '    command: ' + command

    job_config = yacron.config.parse_config_string('''
jobs:
  - name: test
{command}
    schedule: "* * * * *"
    shell: {shell}
    captureStderr: true
    captureStdout: true
    environment:
      - key: FOO
        value: bar
'''.format(command=command_snippet, shell=shell))[0]

    job = yacron.job.RunningJob(job_config, None)

    async def run(job):
        await job.start()
        await job.wait()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run(job))

    if shell_commands:
        run_type = 'shell'
        assert len(shell_commands) == 1
        args, kwargs = shell_commands[0]
    elif exec_commands:
        run_type = 'exec'
        assert len(exec_commands) == 1
        args, kwargs = exec_commands[0]
    else:
        raise AssertionError

    assert kwargs['env']['FOO'] == 'bar'
    assert run_type == expected_type
    assert args == expected_args


def test_execution_timeout():
    job_config = yacron.config.parse_config_string('''
jobs:
  - name: test
    command: |
        echo "hello"
        sleep 1
        echo "world"
    executionTimeout: 0.25
    schedule: "* * * * *"
    captureStderr: false
    captureStdout: true
''')[0]

    async def test(job):
        await job.start()
        await job.wait()
        return job.stdout

    job = yacron.job.RunningJob(job_config, None)
    loop = asyncio.get_event_loop()
    stdout = loop.run_until_complete(test(job))
    assert stdout == "hello\n"


def test_error1():
    job_config = yacron.config.parse_config_string('''
jobs:
  - name: test
    command: echo "hello"
    schedule: "* * * * *"
''')[0]
    job = yacron.job.RunningJob(job_config, None)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(job.start())
    with pytest.raises(RuntimeError):
        loop.run_until_complete(job.start())


def test_error2():
    job_config = yacron.config.parse_config_string('''
jobs:
  - name: test
    command: echo "hello"
    schedule: "* * * * *"
''')[0]
    job = yacron.job.RunningJob(job_config, None)

    loop = asyncio.get_event_loop()
    with pytest.raises(RuntimeError):
        loop.run_until_complete(job.wait())


def test_error3():
    job_config = yacron.config.parse_config_string('''
jobs:
  - name: test
    command: echo "hello"
    schedule: "* * * * *"
''')[0]
    job = yacron.job.RunningJob(job_config, None)

    loop = asyncio.get_event_loop()
    with pytest.raises(RuntimeError):
        loop.run_until_complete(job.cancel())



import asyncio
import asyncio.subprocess
import datetime
import logging
from collections import OrderedDict, defaultdict
from typing import Any, Awaitable, Dict, List, Optional  # noqa

from yacron.config import (JobConfig, parse_config, ConfigError,
                           parse_config_string)
from yacron.job import RunningJob, JobRetryState


logger = logging.getLogger('yacron')
WAKEUP_INTERVAL = datetime.timedelta(minutes=1)


def get_now() -> datetime.datetime:
    return datetime.datetime.utcnow()  # pragma: no cover


def next_sleep_interval() -> float:
    now = get_now()
    target = (datetime.datetime(now.year, now.month, now.day,
                                now.hour, now.minute) +
              WAKEUP_INTERVAL)
    return (target - now).total_seconds()


def create_task(coro: Awaitable) -> asyncio.Task:
    return asyncio.get_event_loop().create_task(coro)


class Cron:

    def __init__(self, config_arg: Optional[str], *,
                 config_yaml: Optional[str]=None
                 ) -> None:
        # list of cron jobs we /want/ to run
        self.cron_jobs = OrderedDict()  # type: Dict[str, JobConfig]
        # list of cron jobs already running
        # name -> list of RunningJob
        self.running_jobs = \
            defaultdict(list)  # type: Dict[str, List[RunningJob]]
        self.config_arg = config_arg
        if config_arg is not None:
            self.update_config()
        if config_yaml is not None:
            # config_yaml is for unit testing
            config = parse_config_string(config_yaml)
            self.cron_jobs = OrderedDict((job.name, job) for job in config)

        self._wait_for_running_jobs_task = None  # type: Optional[asyncio.Task]
        self._stop_event = asyncio.Event()
        self._jobs_running = asyncio.Event()
        self.retry_state = {}  # type: Dict[str, JobRetryState]

    async def run(self) -> None:
        self._wait_for_running_jobs_task = \
            create_task(self._wait_for_running_jobs())

        while not self._stop_event.is_set():
            try:
                self.update_config()
            except ConfigError as err:
                logger.error("Error in configuration file(s), so not updating "
                             "any of the config.:\n%s", str(err))
            except Exception as exc:  # pragma: nocover
                logger.exception("please report this as a bug (1)")
            await self.spawn_jobs()
            sleep_interval = next_sleep_interval()
            logger.debug("Will sleep for %.1f seconds", sleep_interval)
            try:
                await asyncio.wait_for(self._stop_event.wait(), sleep_interval)
            except asyncio.TimeoutError:
                pass

        logger.info("Shutting down (after currently running jobs finish)...")
        while self.retry_state:
            cancel_all = [self.cancel_job_retries(name)
                          for name in self.retry_state]
            await asyncio.gather(*cancel_all)
        await self._wait_for_running_jobs_task

    def signal_shutdown(self) -> None:
        logger.debug("Signalling shutdown")
        self._stop_event.set()

    def update_config(self) -> None:
        if self.config_arg is None:
            return
        config = parse_config(self.config_arg)
        self.cron_jobs = OrderedDict((job.name, job) for job in config)

    async def spawn_jobs(self) -> None:
        now = get_now()
        for job in self.cron_jobs.values():
            if job.schedule.test(now):
                logger.debug("Job %s (%s) is scheduled for now",
                             job.name, job.schedule_unparsed)
                await self.launch_scheduled_job(job)
            else:
                logger.debug("Job %s (%s) not scheduled for now",
                             job.name, job.schedule_unparsed)

    async def launch_scheduled_job(self, job: JobConfig) -> None:
        await self.cancel_job_retries(job.name)
        assert job.name not in self.retry_state

        retry = job.onFailure['retry']
        logger.debug("Job %s retry config: %s", job.name, retry)
        if retry['maximumRetries']:
            retry_state = JobRetryState(retry['initialDelay'],
                                        retry['backoffMultiplier'],
                                        retry['maximumDelay'])
            self.retry_state[job.name] = retry_state

        await self.maybe_launch_job(job)

    async def maybe_launch_job(self, job: JobConfig) -> None:
        if self.running_jobs[job.name]:
            logger.warning("Job %s: still running and concurrencyPolicy is %s",
                           job.name, job.concurrencyPolicy)
            if job.concurrencyPolicy == 'Allow':
                pass
            elif job.concurrencyPolicy == 'Forbid':
                return
            elif job.concurrencyPolicy == 'Replace':
                for running_job in self.running_jobs[job.name]:
                    await running_job.cancel()
            else:
                raise AssertionError  # pragma: no cover
        logger.info("Starting job %s", job.name)
        running_job = RunningJob(job, self.retry_state.get(job.name))
        await running_job.start()
        self.running_jobs[job.name].append(running_job)
        logger.info("Job %s spawned", job.name)
        self._jobs_running.set()

    # continually watches for the running jobs, clean them up when they exit
    async def _wait_for_running_jobs(self) -> None:
        # job -> wait task
        wait_tasks = {}  # type: Dict[RunningJob, asyncio.Task]
        while self.running_jobs or not self._stop_event.is_set():
            try:
                for jobs in self.running_jobs.values():
                    for job in jobs:
                        if job not in wait_tasks:
                            wait_tasks[job] = create_task(job.wait())
                if not wait_tasks:
                    try:
                        await asyncio.wait_for(self._jobs_running.wait(), 1)
                    except asyncio.TimeoutError:
                        pass
                    continue
                self._jobs_running.clear()
                # wait for at least one task with timeout
                done_tasks, _ = await asyncio.wait(
                    wait_tasks.values(),
                    timeout=1.0,
                    return_when=asyncio.FIRST_COMPLETED)
                done_jobs = set()
                for job, task in list(wait_tasks.items()):
                    if task in done_tasks:
                        done_jobs.add(job)
                for job in done_jobs:
                    task = wait_tasks.pop(job)
                    try:
                        task.result()
                    except Exception:  # pragma: no cover
                        logger.exception("please report this as a bug (2)")

                    jobs_list = self.running_jobs[job.config.name]
                    jobs_list.remove(job)
                    if not jobs_list:
                        del self.running_jobs[job.config.name]

                    logger.info("Job %s exit code %s; has stdout: %s, "
                                "has stderr: %s; failed: %s",
                                job.config.name, job.retcode,
                                str(bool(job.stdout)).lower(),
                                str(bool(job.stderr)).lower(),
                                str(job.failed).lower())
                    if job.failed:
                        await self.handle_job_failure(job)
                    else:
                        await self.handle_job_success(job)
            except Exception:  # pragma: no cover
                logger.exception("please report this as a bug (3)")
                await asyncio.sleep(1)

    async def handle_job_failure(self, job: RunningJob) -> None:
        if self._stop_event.is_set():
            return
        if job.stdout:
            logger.info("Job %s STDOUT:\n%s",
                        job.config.name, job.stdout.rstrip())
        if job.stderr:
            logger.info("Job %s STDERR:\n%s",
                        job.config.name, job.stderr.rstrip())
        await job.report_failure()

        # Handle retries...
        state = job.retry_state
        if state is None or state.cancelled:
            await job.report_permanent_failure()
            return

        logger.debug("Job %s has been retried %i times",
                     job.config.name, state.count)
        if state.task is not None:
            if state.task.done():
                await state.task
            else:
                state.task.cancel()
        retry = job.config.onFailure['retry']
        if state.count >= retry['maximumRetries']:
            await self.cancel_job_retries(job.config.name)
            await job.report_permanent_failure()
        else:
            retry_delay = state.next_delay()
            state.task = create_task(
                self.schedule_retry_job(job.config.name, retry_delay,
                                        state.count))

    async def schedule_retry_job(self, job_name: str, delay: float,
                                 retry_num: int) -> None:
        logger.info("Cron job %s scheduled to be retried (#%i) "
                    "in %.1f seconds",
                    job_name, retry_num, delay)
        await asyncio.sleep(delay)
        try:
            job = self.cron_jobs[job_name]
        except KeyError:
            logger.warning("Cron job %s was scheduled for retry, but "
                           "disappeared from the configuration", job_name)
        await self.maybe_launch_job(job)

    async def handle_job_success(self, job: RunningJob) -> None:
        await self.cancel_job_retries(job.config.name)
        await job.report_success()

    async def cancel_job_retries(self, name: str) -> None:
        try:
            state = self.retry_state.pop(name)
        except KeyError:
            return
        state.cancelled = True
        if state.task is not None:
            if state.task.done():
                await state.task
            else:
                state.task.cancel()



import asyncio
import asyncio.subprocess
import logging
import os
import sys
import time
from email.mime.text import MIMEText
from typing import Any, Awaitable, Dict, List, Optional  # noqa
from enum import Enum

from raven import Client
from raven_aiohttp import AioHttpTransport

import aiosmtplib
from yacron.config import JobConfig

logger = logging.getLogger('yacron')


def create_task(coro: Awaitable) -> asyncio.Task:
    return asyncio.get_event_loop().create_task(coro)


class ReportType(Enum):
    FAILURE = 1
    SUCCESS = 2


class StreamReader:

    def __init__(self,
                 job_name: str,
                 stream_name: str,
                 stream: asyncio.StreamReader,
                 save_limit: int) -> None:
        self.save_top = []  # type: List[str]
        self.save_bottom = []  # type: List[str]
        self.job_name = job_name
        self.save_limit = save_limit
        self.stream_name = stream_name
        self._reader = create_task(self._read(stream))
        self.discarded_lines = 0

    async def _read(self, stream):
        prefix = "[{} {}] ".format(self.job_name, self.stream_name)
        limit_top = self.save_limit // 2
        limit_bottom = self.save_limit - limit_top
        while True:
            line = (await stream.readline()).decode("utf-8")
            if not line:
                return
            sys.stdout.write(prefix + line)
            sys.stdout.flush()
            if len(self.save_top) < limit_top:
                self.save_top.append(line)
            else:
                if len(self.save_bottom) == limit_bottom:
                    del self.save_bottom[0]
                    self.discarded_lines += 1
                self.save_bottom.append(line)

    async def join(self) -> str:
        await self._reader
        if self.save_bottom:
            middle = (["   [.... {} lines discarded ...]\n"
                       .format(self.discarded_lines)]
                      if self.discarded_lines else [])
            output = ''.join(self.save_top + middle + self.save_bottom)
        else:
            output = ''.join(self.save_top)
        return output


class Reporter:

    async def report(self, report_type: ReportType, job: 'RunningJob',
                     config: Dict[str, Any]) -> None:
        raise NotImplementedError  # pragma: no cover

    @staticmethod
    def _format_body(job):
        if job.stdout and job.stderr:
            body = ("STDOUT:\n---\n{}\n---\nSTDERR:\n{}"
                    .format(job.stdout, job.stderr))
        else:
            body = job.stdout or job.stderr or '(no output was captured)'
        return body


class SentryReporter(Reporter):

    async def report(self, report_type: ReportType, job: 'RunningJob',
                     config: Dict[str, Any]) -> None:
        config = config['sentry']
        if config['dsn']['value']:
            dsn = config['dsn']['value']
        elif config['dsn']['fromFile']:
            with open(config['dsn']['fromFile'], "rt") as dsn_file:
                dsn = dsn_file.read().strip()
        elif config['dsn']['fromEnvVar']:
            dsn = os.environ[config['dsn']['fromEnvVar']]
        else:
            return  # sentry disabled: early return

        body = self._format_body(job)

        if report_type == ReportType.SUCCESS:
            headline = ('Cron job {!r} completed'
                        .format(job.config.name))
        elif report_type == ReportType.FAILURE:
            headline = ('Cron job {!r} failed'
                        .format(job.config.name))
        body = "{}\n\n{}".format(headline, body)

        client = Client(transport=AioHttpTransport,
                        dsn=dsn,
                        string_max_length=4096)
        extra = {
            'job': job.config.name,
            'exit_code': job.retcode,
            'command': job.config.command,
            'shell': job.config.shell,
            'success': True if report_type == ReportType.SUCCESS else False,
        }
        logger.debug("sentry body: %r", body)
        client.captureMessage(
            body,
            extra=extra,
        )


class MailReporter(Reporter):

    async def report(self, report_type: ReportType, job: 'RunningJob',
                     config: Dict[str, Any]) -> None:
        mail = config['mail']
        if not ((mail['smtpHost'] or mail['smtp_host']) and
                mail['to'] and mail['from']):
            return  # email reporting disabled

        body = self._format_body(job)

        if mail['smtpHost']:
            smtp_host = mail['smtpHost']
        else:  # pragma: no cover
            logger.warning("smtp_host is deprecated, was renamed to smtpHost")
            smtp_host = mail['smtp_host']
        if mail['smtpPort']:
            smtp_port = mail['smtpPort']
        else:  # pragma: no cover
            logger.warning("smtp_port is deprecated, was renamed to smtpPort")
            smtp_port = mail['smtp_port']

        logger.debug("smtp: host=%r, port=%r", smtp_host, smtp_port)
        smtp = aiosmtplib.SMTP(hostname=smtp_host, port=smtp_port)
        await smtp.connect()
        message = MIMEText(body)
        message['From'] = mail['from']
        message['To'] = mail['to']
        if report_type == ReportType.SUCCESS:
            message['Subject'] = ('Cron job {!r} completed'
                                  .format(job.config.name))
        elif report_type == ReportType.FAILURE:
            message['Subject'] = ('Cron job {!r} failed'
                                  .format(job.config.name))
        else:  # pragma: no cover
            raise AssertionError
        await smtp.send_message(message)


class JobRetryState:

    def __init__(self, initial_delay: float,
                 multiplier: float,
                 max_delay: float) -> None:
        self.multiplier = multiplier
        self.max_delay = max_delay
        self.delay = initial_delay
        self.count = 0  # number of times retried
        self.task = None  # type: Optional[asyncio.Task]
        self.cancelled = False

    def next_delay(self) -> float:
        delay = self.delay
        self.delay = min(delay * self.multiplier, self.max_delay)
        self.count += 1
        return delay


class RunningJob:
    REPORTERS = [
        SentryReporter(),
        MailReporter(),
    ]  # type: List[Reporter]

    def __init__(self, config: JobConfig,
                 retry_state: Optional[JobRetryState]) -> None:
        self.config = config
        self.proc = None  # type: Optional[asyncio.subprocess.Process]
        self.retcode = None  # type: Optional[int]
        self._stderr_reader = None  # type: Optional[StreamReader]
        self._stdout_reader = None  # type: Optional[StreamReader]
        self.stderr = None  # type: Optional[str]
        self.stdout = None  # type: Optional[str]
        self.execution_deadline = None  # type: Optional[float]
        self.retry_state = retry_state

    async def start(self) -> None:
        if self.proc is not None:
            raise RuntimeError("process already running")
        kwargs = {}  # type: Dict[str, Any]
        if isinstance(self.config.command, list):
            create = asyncio.create_subprocess_exec
            cmd = self.config.command
        else:
            if self.config.shell:
                create = asyncio.create_subprocess_exec
                cmd = [self.config.shell, '-c', self.config.command]
            else:
                create = asyncio.create_subprocess_shell
                cmd = [self.config.command]
        if self.config.environment:
            env = dict(os.environ)
            for envvar in self.config.environment:
                env[envvar['key']] = envvar['value']
            kwargs['env'] = env
        logger.debug("%s: will execute argv %r", self.config.name, cmd)
        if self.config.captureStderr:
            kwargs['stderr'] = asyncio.subprocess.PIPE
        if self.config.captureStdout:
            kwargs['stdout'] = asyncio.subprocess.PIPE
        if self.config.executionTimeout:
            self.execution_deadline = (time.perf_counter() +
                                       self.config.executionTimeout)

        self.proc = await create(*cmd, **kwargs)

        if self.config.captureStderr:
            assert self.proc.stderr is not None
            self._stderr_reader = \
                StreamReader(self.config.name, 'stderr', self.proc.stderr,
                             self.config.saveLimit)
        if self.config.captureStdout:
            assert self.proc.stdout is not None
            self._stdout_reader = \
                StreamReader(self.config.name, 'stdout', self.proc.stdout,
                             self.config.saveLimit)

    async def wait(self) -> None:
        if self.proc is None:
            raise RuntimeError("process is not running")
        if self.execution_deadline is None:
            self.retcode = await self.proc.wait()
        else:
            timeout = self.execution_deadline - time.perf_counter()
            try:
                if timeout > 0:
                    self.retcode = await asyncio.wait_for(
                        self.proc.wait(),
                        timeout,
                    )
                else:
                    raise asyncio.TimeoutError
            except asyncio.TimeoutError:
                logger.info("Job %s exceeded its executionTimeout of "
                            "%.1f seconds, cancelling it...",
                            self.config.name, self.config.executionTimeout)
                self.retcode = -100
                await self.cancel()
        if self._stderr_reader:
            self.stderr = await self._stderr_reader.join()
        if self._stdout_reader:
            self.stdout = await self._stdout_reader.join()

    @property
    def failed(self) -> bool:
        if self.config.failsWhen['nonzeroReturn'] and self.retcode != 0:
            return True
        if self.config.failsWhen['producesStdout'] and self.stdout:
            return True
        if self.config.failsWhen['producesStderr'] and self.stderr:
            return True
        return False

    async def cancel(self) -> None:
        if self.proc is None:
            raise RuntimeError("process is not running")
        self.proc.terminate()
        try:
            await asyncio.wait_for(self.proc.wait(), self.config.killTimeout)
        except asyncio.TimeoutError:
            logger.warning("Job %s did not gracefully terminate after "
                           "%.1f seconds, killing it...",
                           self.config.name, self.config.killTimeout)
            self.proc.kill()

    async def report_failure(self):
        logger.info("Cron job %s: reporting failure", self.config.name)
        await self._report_common(self.config.onFailure['report'],
                                  ReportType.FAILURE)

    async def report_permanent_failure(self):
        logger.info("Cron job %s: reporting permanent failure",
                    self.config.name)
        await self._report_common(self.config.onPermanentFailure['report'],
                                  ReportType.FAILURE)

    async def report_success(self):
        logger.info("Cron job %s: reporting success", self.config.name)
        await self._report_common(self.config.onSuccess['report'],
                                  ReportType.SUCCESS)

    async def _report_common(self, report_config: dict,
                             report_type: ReportType) -> None:
        results = await asyncio.gather(
            *[reporter.report(report_type, self, report_config)
              for reporter in self.REPORTERS],
            return_exceptions=True
        )
        for result in results:
            if isinstance(result, Exception):
                logger.error("Problem reporting job %s failure: %s",
                             self.config.name, result)
